package se.rupy.http;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;

/**
 * Hot-deploys an application containing one or many service filters from disk
 * with simplistic dynamic class loading, eventually after receiving it through
 * a HTTP POST.
 * @author marc
 */
public class Deploy extends Service {
	public static String path, pass;

	public Deploy(String path, String pass) {
		Deploy.path = path;
		Deploy.pass = pass;

		new File(path).mkdirs();
	}

	public String path() {
		return "/deploy";
	}

	public void filter(Event event) throws Event, Exception {
		String name = event.query().header("file");
		String pass = event.query().header("pass");

		if (name == null) {
			throw new Failure("File header missing.");
		}

		if (pass == null) {
			throw new Failure("Pass header missing.");
		} else if (!Deploy.pass.equals(pass)) {
			throw new Failure("Pass verification failed. (" + pass + ")");
		}

		File file = new File(path + name);
		OutputStream out = new FileOutputStream(file);
		InputStream in = event.query().input();

		pipe(in, out, 10240);

		out.flush();
		out.close();

		String deploy = deploy(event.daemon(), file);

		event.reply().output().print("Application '" + deploy + "' deployed.");
	}

	public static String deploy(Daemon daemon, File file) throws Exception {
		Archive archive = new Archive(daemon, file);

		daemon.chain(archive);
		daemon.verify(archive);

		return archive.name();
	}

	static class Archive extends ClassLoader {
		private HashSet service;
		//private HashMap content;
		private HashMap chain;
		private String name;
		private String host;
		private long date;

		Archive(Daemon daemon, File file) throws Exception {
			service = new HashSet();
			//content = new HashMap();
			chain = new HashMap();
			name = file.getName();
			date = file.lastModified();

			JarInputStream in = new JarInput(new FileInputStream(file));
			Attributes attr = in.getManifest().getMainAttributes();
			
			host = (String) attr.get("host");
			
			if(host == null) {
				host = "content";
			}
			
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			JarEntry entry = null;

			Vector classes = new Vector();

			while ((entry = in.getNextJarEntry()) != null) {
				if (entry.getName().endsWith(".class")) {
					pipe(in, out);
					byte[] data = out.toByteArray();
					out.reset();

					String name = name(entry.getName());
					classes.add(new Small(name, data));
				} else if (!entry.isDirectory()) {
					Big.write(host, "/" + entry.getName(), in);
					//content.put("/" + entry.getName(), new Big(host, "/" + entry.getName(), in, date));
				}
			}

			int length = classes.size();
			String missing = "";
			Small small = null;

			while (classes.size() > 0) {
				try {
					small = (Small) classes.elementAt(0);
					classes.removeElement(small);
					instantiate(small);
				} catch (NoClassDefFoundError e) {
					// the superclass has not been loaded yet
					if(!missing.equals(e.getMessage())) {
						missing = e.getMessage();
						length = classes.size();
					}
					classes.addElement(small);
					length--;
					if (length < 0) {
						throw e;
					}
				}
			}
		}

		void instantiate(Small small) throws Exception {
			if (small.clazz == null) {
				small.clazz = defineClass(small.name, small.data, 0,
						small.data.length);
			}

			Class clazz = small.clazz.getSuperclass();
			boolean service = false;
			
			while (clazz != null) {
				if (clazz.getCanonicalName().equals("se.rupy.http.Service")) {
					service = true;
				}
				clazz = clazz.getSuperclass();
			}
			
			if(service) {
				try {
					this.service.add((Service) small.clazz.newInstance());
				}
				catch(InstantiationException e) {
					// abstract superclass ...
				}
			}
		}

		static String name(String name) {
			name = name.substring(0, name.indexOf("."));
			name = name.replace("/", ".");
			return name;
		}

		public String name() {
			return name;
		}

		public String host() {
			return host;
		}
		
		public long date() {
			return date;
		}

		public HashMap chain() {
			return chain;
		}
/*
		public HashMap content() {
			return content;
		}
*/
		public HashSet service() {
			return service;
		}
	}

	static class Big implements Stream {
		private File file;
		private String name;
		private long date;

		public Big(String host, String name, InputStream in, long date) throws IOException {
			file = write(host, name, in);

			this.name = name;
			this.date = date - date % 1000;
		}

		public Big(File file) {
			long date = file.lastModified();
			this.name = file.getName();
			this.file = file;
			this.date = date - date % 1000;
		}
		
		static File write(String host, String name, InputStream in) throws IOException {
			String path = name.substring(0, name.lastIndexOf("/"));
			String root = Deploy.path + host;

			new File(root + path).mkdirs();

			File file = new File(root + name);
			file.createNewFile();

			OutputStream out = new FileOutputStream(file);
			
			pipe(in, out);
			
			out.flush();
			out.close();
			
			return file;
		}
		
		public String name() {
			return name;
		}

		public InputStream input() {
			try {
				return new FileInputStream(file);
			} catch (FileNotFoundException e) {
				return null;
			}
		}

		public long length() {
			return file.length();
		}

		public long date() {
			return date;
		}
	}

	static class Small implements Stream {
		private String name;
		private byte[] data;
		private long date;
		private Class clazz;

		public Small(String name, byte[] data) {
			this(name, data, 0);
		}

		public Small(String name, byte[] data, long date) {
			this.name = name;
			this.data = data;
			this.date = date - date % 1000;
		}

		public String name() {
			return name;
		}

		public InputStream input() {
			return new ByteArrayInputStream(data);
		}

		public long length() {
			return data.length;
		}

		public long date() {
			return date;
		}
		
		byte[] data() {
			return data;
		}
	}

	static interface Stream {
		public String name();
		public InputStream input();
		public long length();
		public long date();
	}

	static class Client {
		InputStream send(URL url, File file, String pass) throws IOException {
			HttpURLConnection conn = (HttpURLConnection) url.openConnection();
			conn.setRequestMethod("POST");

			OutputStream out = null;
			InputStream in = null;

			if (file != null) {
				conn.addRequestProperty("File", file.getName());

				if (pass != null) {
					conn.addRequestProperty("Pass", pass);
				}

				conn.setChunkedStreamingMode(0);
				conn.setDoOutput(true);

				out = conn.getOutputStream();
				in = new FileInputStream(file);

				pipe(in, out);

				out.flush();
				in.close();
			}

			int code = conn.getResponseCode();

			if (code == 200) {
				in = conn.getInputStream();
			} else if (code < 0) {
				throw new IOException("HTTP response unreadable.");
			} else {
				in = conn.getErrorStream();
			}

			return in;
		}

		static String toString(InputStream in) throws IOException {
			ByteArrayOutputStream out = new ByteArrayOutputStream();

			pipe(in, out);

			out.close();
			in.close();

			return new String(out.toByteArray());
		}
	}

	public static String name(String name) {
		name = name.substring(0, name.indexOf("."));
		name = name.replace("/", ".");
		return name;
	}

	public static int pipe(InputStream in, OutputStream out) throws IOException {
		return pipe(in, out, 1024, 0);
	}

	public static int pipe(InputStream in, OutputStream out, int length)
	throws IOException {
		return pipe(in, out, length, 0);
	}

	public static int pipe(InputStream in, OutputStream out, int length,
			int limit) throws IOException {
		byte[] data = new byte[length];
		int total = 0, read = in.read(data);
		while (read > -1) {
			if (limit > 0 && total > limit) {
				throw new IOException("Max allowed bytes read. (" + limit + ")");
			}
			total += read;
			out.write(data, 0, read);
			read = in.read(data);
		}
		return total;
	}

	/**
	 * Avoids the jar stream being cutoff.
	 * @author marc.larue
	 */
	static class JarInput extends JarInputStream {
		public JarInput(InputStream in) throws IOException {
			super(in);
		}

		public void close() {
			// geez
		}
	}

	/**
	 * <pre>
	 * &lt;target name="deploy"&gt;
	 * &lt;java fork="yes" 
	 *     classname="se.rupy.http.Deploy" 
	 *     classpath="http.jar"&gt;
	 *      &lt;arg line="localhost:8000"/&gt;&lt;!-- any host:port --&gt;
	 *      &lt;arg line="service.jar"/&gt;&lt;!-- your application jar --&gt;
	 *      &lt;arg line="secret"/&gt;&lt;!-- see run.bat and run.sh --&gt;
	 * &lt;/java&gt;
	 * &lt;/target&gt;
	 * </pre>
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length > 2) {
			try {
				URL url = new URL("http://" + args[0] + "/deploy");
				File file = new File(args[1]);
				InputStream in = new Client().send(url, file, args[2]);
				System.out.println(Client.toString(in));
			} catch (ConnectException ce) {
				System.out
				.println("Connection failed, is there a server running on "
						+ args[0] + "?");
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			System.out.println("Usage: Deploy [host] [file] [pass]");
		}
	}
}
